import os
from importlib import import_module
from arkitekt.apps import App
from typing import Type


def build_relative_dir(*paths):
    return os.path.join(os.path.dirname(os.path.abspath(__file__)), *paths)


def import_builder(builder: str) -> Type[App]:
    module_path, function_name = builder.rsplit(".", 1)
    module = import_module(module_path)
    function = getattr(module, function_name)
    return function



from arkitekt.apps.fakts import ArkitektFakts
from arkitekt.apps.herre import ArkitektHerre
from arkitekt.apps.mikro import ArkitektMikro
from arkitekt.apps.fluss import ArkitektFluss
from arkitekt.apps.rekuest import ArkitektRekuest
from arkitekt.apps.unlok import ArkitektUnlok
from koil.composition import Composition
from arkitekt.apps.fakts import build_arkitekt_fakts, build_arkitekt_token_fakts
from arkitekt.apps.herre import build_arkitekt_herre
from arkitekt.apps.rekuest import build_arkitekt_rekuest
from arkitekt.apps.mikro import build_arkitekt_mikro
from arkitekt.apps.unlok import build_arkitekt_unlok
from arkitekt.apps.fluss import build_arkitekt_fluss
from arkitekt.model import Manifest


class App(Composition):
    manifest: Manifest
    fakts: ArkitektFakts
    herre: ArkitektHerre
    rekuest: ArkitektRekuest
    mikro: ArkitektMikro
    unlok: ArkitektUnlok
    fluss: ArkitektFluss
    """
    Arkitekt

    An app that connected to the services of the arkitekt Api,
    it comes included with the following services:

    - Rekuest: A service for that handles requests to the arkitekt Api as well as provides an
      interface to provide functionality on the arkitekt Api.
    - Herre: A service for that handles the authentication and authorization of the user
    - Fakts: A service for that handles the discovery
      and retrieval of the configuration of the arkitekt Api
    - Mikro: A service for that handles the storage and data of microscopy data

    Apps have to be always used within a context manager, this is to ensure that the services
    are properly closed when the app is no longer needed.

    Example:
        >>> from arkitekt import Arkitekt
        >>> app = Arkitekt()
        >>> with app:
        >>>     # Do stuff
        >>> # App is closed

    """


def build_arkitekt_app(
    manifest: Manifest,
    url=None,
    no_cache=False,
    headless=False,
    instance_id=None,
    token=None,
):
    fakts = (
        build_arkitekt_fakts(
            manifest=manifest, url=url, no_cache=no_cache, headless=headless
        )
        if not token
        else build_arkitekt_token_fakts(
            manifest=manifest,
            token=token,
            url=url,
            no_cache=no_cache,
            headless=headless,
        )
    )
    herre = build_arkitekt_herre(
        manifest=manifest, fakts=fakts, url=url, no_cache=no_cache
    )
    rekuest = build_arkitekt_rekuest(fakts=fakts, herre=herre, instance_id=instance_id)
    mikro = build_arkitekt_mikro(fakts=fakts, herre=herre)
    unlok = build_arkitekt_unlok(herre=herre, fakts=fakts)
    fluss = build_arkitekt_fluss(herre=herre, fakts=fakts)

    return App(
        manifest=manifest,
        fakts=fakts,
        herre=herre,
        rekuest=rekuest,
        mikro=mikro,
        unlok=unlok,
        fluss=fluss,
    )



from arkitekt.apps.fakts_qt import ArkitektFaktsQt, build_arkitekt_qt_fakts
from arkitekt.apps.herre_qt import ArkitektHerreQt, build_arkitekt_qt_herre
from arkitekt.apps.mikro import ArkitektMikro
from arkitekt.apps.rekuest import ArkitektRekuest
from arkitekt.apps.unlok import ArkitektUnlok
from arkitekt.apps.fluss import ArkitektFluss
from koil.composition import Composition
from arkitekt.model import Manifest
from arkitekt.apps.mikro import build_arkitekt_mikro
from arkitekt.apps.unlok import build_arkitekt_unlok
from arkitekt.apps.fluss import build_arkitekt_fluss
from arkitekt.apps.reaktion_rekuest import build_arkitekt_reaktion_rekuest


class QtSchedulingApp(Composition):
    manifest: Manifest
    fakts: ArkitektFaktsQt
    herre: ArkitektHerreQt
    rekuest: ArkitektRekuest
    mikro: ArkitektMikro
    unlok: ArkitektUnlok
    fluss: ArkitektFluss
    """
    Arkitekt

    An app that connected to the services of the arkitekt Api,
    it comes included with the following services:

    - Rekuest: A service for that handles requests to the arkitekt Api as well as provides an
      interface to provide functionality on the arkitekt Api.
    - Herre: A service for that handles the authentication and authorization of the user
    - Fakts: A service for that handles the discovery
      and retrieval of the configuration of the arkitekt Api
    - Mikro: A service for that handles the storage and data of microscopy data

    Apps have to be always used within a context manager, this is to ensure that the services
    are properly closed when the app is no longer needed.

    Example:
        >>> from arkitekt import Arkitekt
        >>> app = Arkitekt()
        >>> with app:
        >>>     # Do stuff
        >>> # App is closed

    """


def build_arkitekt_scheduleqt_app(
    manifest: Manifest,
    no_cache=False,
    instance_id=None,
    beacon_widget=None,
    login_widget=None,
    parent=None,
    settings=None,
):
    fakts = build_arkitekt_qt_fakts(
        manifest=manifest,
        no_cache=no_cache,
        beacon_widget=beacon_widget,
        parent=parent,
        settings=settings,
    )
    herre = build_arkitekt_qt_herre(
        manifest=manifest,
        fakts=fakts,
        login_widget=login_widget,
        parent=parent,
        settings=settings,
    )
    rekuest = build_arkitekt_reaktion_rekuest(
        fakts=fakts, herre=herre, instance_id=instance_id
    )
    mikro = build_arkitekt_mikro(fakts=fakts, herre=herre)
    unlok = build_arkitekt_unlok(herre=herre, fakts=fakts)
    fluss = build_arkitekt_fluss(herre=herre, fakts=fakts)

    return QtSchedulingApp(
        manifest=manifest,
        fakts=fakts,
        herre=herre,
        rekuest=rekuest,
        mikro=mikro,
        unlok=unlok,
        fluss=fluss,
    )



from ssl import SSLContext
import ssl
from typing import Any, Dict, Optional

import aiohttp
import certifi
from pydantic import Field
from koil.composition import KoiledModel
from fakts.fakt.base import Fakt
from fakts.fakts import Fakts


class HealthzConfig(Fakt):
    healthz: str


class FaktsHealthz(KoiledModel):
    ssl_context: SSLContext = Field(
        default_factory=lambda: ssl.create_default_context(cafile=certifi.where())
    )
    fakts: Fakts
    fakts_group: str
    strict: bool = False
    endpoint_url: Optional[str]
    fakt: Optional[HealthzConfig]

    _old_fakt: Dict[str, Any] = None

    def configure(self, config: HealthzConfig):
        self.endpoint_url = config.healthz

    async def check(self):
        self.endpoint_url = await self.fakts.aget(self.fakts_group)["healthz"]

        async with aiohttp.ClientSession(
            headers={"Content-Type": "application/json"},
            connector=aiohttp.TCPConnector(ssl=self.ssl_context),
        ) as session:
            # get json from endpoint
            async with session.get(self.endpoint_url + "?format=json") as resp:
                healthz_json = await resp.json()
                if self.strict:
                    faulty_services = []
                    for key, value in healthz_json.items():
                        if value != "working":
                            faulty_services.append({"service": key, "error": value})

                    if faulty_services:
                        raise Exception(f"Faulty services: {faulty_services}")

                if "detail" in healthz_json:
                    raise Exception("Error on the Healtzh endpoint ")

                return healthz_json

    class Config:
        underscore_attrs_are_private = True



from importlib import import_module
from rich.console import Console
import asyncio
from .ui import construct_run_panel
from .utils import import_builder
from .types import Manifest

console = Console()


async def run_app(entrypoint, app):
    try:
        with console.status("Loading entrypoint module..."):
            try:
                import_module(entrypoint)
            except ModuleNotFoundError as e:
                console.print(f"Could not find entrypoint module {entrypoint}")
                raise e

        console.print("App is running...")
    except asyncio.CancelledError as e:
        console.print("Shutting down...")
        raise e


def import_entrypoint(entrypoint):
    with console.status("Loading entrypoint module..."):
        try:
            import_module(entrypoint)
        except ModuleNotFoundError as e:
            console.print(f"Could not find entrypoint module {entrypoint}")
            raise e


async def run_production(
    manifest: Manifest,
    entrypoint=None,
    builder: str = None,
    url=None,
    version=None,
    **builder_kwargs,
):
    builder = import_builder(builder)

    import_entrypoint(entrypoint or manifest.entrypoint)

    app = builder(
        identifier=manifest.identifier,
        version=version or manifest.version,
        logo=manifest.logo,
        url=url,
        **builder_kwargs,
    )

    panel = construct_run_panel(app)
    console.print(panel)

    async with app:
        await app.rekuest.run()



from herre.herre import Herre
from fakts import Fakts
from herre.grants import CacheGrant as HerreCacheGrant
from herre.grants.oauth2.refresh import RefreshGrant
from herre.fakts.fakts_endpoint_fetcher import FaktsUserFetcher
from herre.fakts.grant import FaktsGrant
from typing import Optional
from arkitekt.model import Manifest, User


class ArkitektHerre(Herre):
    pass


def build_arkitekt_herre(
    manifest: Manifest, fakts: Fakts, url: str, no_cache: Optional[bool] = False
) -> ArkitektHerre:
    return ArkitektHerre(
        grant=RefreshGrant(
            grant=FaktsGrant(fakts=fakts, fakts_group="lok"),
        ),
        fetcher=FaktsUserFetcher(
            fakts=fakts, fakts_key="lok.userinfo_url", userModel=User
        ),
    )



from arkitekt.apps.fakts import ArkitektFakts
from arkitekt.apps.herre import ArkitektHerre
from arkitekt.apps.mikro import ArkitektMikro
from arkitekt.apps.fluss import ArkitektFluss
from arkitekt.apps.mikro_next import ArkitektMikroNext
from arkitekt.apps.rekuest_next import ArkitektRekuestNext
from arkitekt.apps.unlok import ArkitektUnlok
from koil.composition import Composition
from fakts.grants.remote import Manifest
from arkitekt.apps.fakts import build_arkitekt_fakts, build_arkitekt_token_fakts
from arkitekt.apps.herre import build_arkitekt_herre
from arkitekt.apps.rekuest_next import build_arkitekt_rekuest_next
from arkitekt.apps.mikro_next import build_arkitekt_mikro_next
from arkitekt.apps.mikro import build_arkitekt_mikro
from arkitekt.apps.unlok import build_arkitekt_unlok
from arkitekt.apps.fluss import build_arkitekt_fluss
from koil.composition import PedanticKoil


class NextApp(Composition):
    manifest: Manifest
    fakts: ArkitektFakts
    herre: ArkitektHerre
    rekuest: ArkitektRekuestNext
    mikro: ArkitektMikro
    mikro_next: ArkitektMikroNext
    unlok: ArkitektUnlok
    fluss: ArkitektFluss
    """
    Arkitekt

    An app that connected to the services of the arkitekt Api,
    it comes included with the following services:

    - Rekuest: A service for that handles requests to the arkitekt Api as well as provides an
      interface to provide functionality on the arkitekt Api.
    - Herre: A service for that handles the authentication and authorization of the user
    - Fakts: A service for that handles the discovery
      and retrieval of the configuration of the arkitekt Api
    - Mikro: A service for that handles the storage and data of microscopy data

    Apps have to be always used within a context manager, this is to ensure that the services
    are properly closed when the app is no longer needed.

    Example:
        >>> from arkitekt import Arkitekt
        >>> app = Arkitekt()
        >>> with app:
        >>>     # Do stuff
        >>> # App is closed

    """


def build_next_app(
    manifest: Manifest,
    url=None,
    no_cache=False,
    headless=False,
    instance_id=None,
    token=None,
):
    fakts = (
        build_arkitekt_fakts(
            manifest=manifest, url=url, no_cache=no_cache, headless=headless
        )
        if not token
        else build_arkitekt_token_fakts(
            manifest=manifest,
            token=token,
            url=url,
            no_cache=no_cache,
            headless=headless,
        )
    )
    herre = build_arkitekt_herre(
        manifest=manifest, fakts=fakts, url=url, no_cache=no_cache
    )
    rekuest = build_arkitekt_rekuest_next(
        fakts=fakts, herre=herre, instance_id=instance_id
    )
    mikro = build_arkitekt_mikro(fakts=fakts, herre=herre)
    mikro_new = build_arkitekt_mikro_next(fakts=fakts, herre=herre)
    unlok = build_arkitekt_unlok(herre=herre, fakts=fakts)
    fluss = build_arkitekt_fluss(herre=herre, fakts=fakts)

    return NextApp(
        manifest=manifest,
        fakts=fakts,
        herre=herre,
        rekuest=rekuest,
        mikro=mikro,
        mikro_next=mikro_new,
        unlok=unlok,
        fluss=fluss,
    )
